<?
require_once("ldbi.php");
require_once("xml.inc");

define("SELECT_CMD", "SELECT ");
define("INSERT_CMD", "INSERT INTO ");
define("UPDATE_CMD", "UPDATE ");
define("CREATE_TABLE_CMD", "CREATE TABLE ");
define("CREATE_DB_CMD", "CREATE DATABASE ");
define("CREATE_INDEX_CMD", "CREATE INDEX ");
define("CREATE_SEQ_CMD", "CREATE SEQUENCE ");

class SQLi
{
	var $dbh;
	var $tables;
	var $sequences;
	var $sqlfunctions;

	function connect($dbname)
	{
		$this->dbh = &new LDBI($dbname);

		if($this->dbh)
		{
			$this->dbh->load_tables();
			if(is_array($this->dbh->tables))
			{
				foreach($this->dbh->tables as $obj)
				{
					$this->load_table_config($obj);
				}
				reset($this->dbh->tables);
			}

			$this->dbh->load_sequences();
			if(is_array($this->dbh->sequences))
			{
				foreach($this->dbh->sequences as $obj)
				{
					$this->load_sequence_config($obj);
				}
				reset($this->dbh->sequences);
			}

			$this->sqlfunctions['nextval'] = array(
				"call"=>"sequence_next",
				"type"=>"int"
			);
			$this->sqlfunctions['currval'] = array(
				"call"=>"sequence_curr",
				"type"=>"int"
			);
			$this->sqlfunctions['setval'] = array(
				"call"=>"sequence_set",
				"type"=>"bool"
			);
		}
	}

	function split_query($query)
	{
		$query = trim($query);
		if(!strncasecmp($query, SELECT_CMD, strlen(SELECT_CMD)))
		{
			$sp = split(" ", $query);
			$res[] = $sp[0];
			$res[] = $sp[1];
			$res[] = $sp[3];

			foreach($sp as $l=>$v)
			{
				if($l < 5) continue;
				$p[] = $v;
			}
			if(is_array($p))
				$res[] = join(" ", $p);
		}
		else if(!strncasecmp($query, CREATE_TABLE_CMD,
				strlen(CREATE_TABLE_CMD)))
		{
			$sp = split(" ", $query);
			$res[] = $sp[0];
			$res[] = $sp[1];
			$res[] = $sp[2];

			foreach($sp as $l=>$v)
			{
				if($l < 3) continue;
				$p[] = $v;
			}
			$tmp = trim(join(" ", $p));
			unset($p);
			if(!strncmp($tmp, "(", 1))
			{
				$tmp = substr($tmp, 1);
			}
			else
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad syntax: $query\n");
				unset($res);
				unset($sp);
				return FALSE;
			}
			if(substr($tmp, -1, 1) == ")")
			{
				$tmp = substr($tmp, 0, -1);
			}
			else
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad syntax: $query\n");
				unset($res);
				unset($sp);
				return FALSE;
			}
			$res[] = trim($tmp);
		}
		else if(!strncasecmp($query, CREATE_DB_CMD,
				strlen(CREATE_DB_CMD)))
		{
			$sp = split(" ", $query);
			$res[] = $sp[0];
			$res[] = $sp[1];
			$res[] = $sp[2];
		}
		else if(!strncasecmp($query, CREATE_INDEX_CMD,
				strlen(CREATE_INDEX_CMD)))
		{
			$sp = split(" ", $query);
			$res[] = $sp[0];
			$res[] = $sp[1];
			$res[] = $sp[2];
			if(strcasecmp($sp[3], "on"))
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: $query\n");
				unset($res);
				unset($sp);
				return FALSE;
			}
			$res[] = $sp[4];

			foreach($sp as $l=>$v)
			{
				if($l < 5) continue;
				$p[] = $v;
			}
			$tmp = trim(join(" ", $p));
			if(!strncmp($tmp, "(", 1))
			{
				$tmp = substr($tmp, 1);
			}
			else
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad syntax: $query\n");
				unset($res);
				unset($sp);
				return FALSE;
			}
			if(substr($tmp, -1, 1) == ")")
			{
				$tmp = substr($tmp, 0, -1);
			}
			else
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad syntax: $query\n");
				unset($res);
				unset($sp);
				return FALSE;
			}
			$res[] = trim($tmp);
			unset($tmp);
		}
		else if(!strncasecmp($query, CREATE_SEQ_CMD,
				strlen(CREATE_SEQ_CMD)))
		{
			$sp = split(" ", "$query");
			$res[] = $sp[0];
			$res[] = $sp[1];
			$res[] = $sp[2];

			for($i = 3 ; $i < count($sp); $i++)
			{
				if(!strcasecmp($sp[$i], "INCREMENT"))
				{
					if(!isset($increment))
					{
						$increment = (int)$sp[$i + 1];
						if(!is_int($increment))
						{
							$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: '$increment' is a bad value for INCREMENT");
							return FALSE;
						}
					}
					else
					{
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: INCREMENT already set");
						return FALSE;
					}
				}
				else if(!strcasecmp($sp[$i], "START"))
				{
					if(!isset($start))
					{
						$start = (int)$sp[$i + 1];
						if(!is_int($start))
						{
							$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: '$start' is a bad value for START");
							return FALSE;
						}
					}
					else
					{
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: START already set");
						return FALSE;
					}
				}
				else if(!strcasecmp($sp[$i], "MAXVALUE"))
				{
					if(!isset($maxval))
					{
						$maxval = (int)$sp[$i + 1];
						if(!is_int($maxval))
						{
							$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: '$maxval' is a bad value for MAXVALUE");
							return FALSE;
						}
					}
					else
					{
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: MAXVALUE already set");
						return FALSE;
					}
				}
			}

			if(isset($start))
				$res[] = $start;
			if(isset($increment))
				$res[] = $increment;
			if(isset($maxval))
				$res[] = $maxval;
		}
		else if(!strncasecmp($query, INSERT_CMD,
				strlen(INSERT_CMD)))
		{
			$sp = split(" ", $query);
			$res[] = $sp[0];
			if(strcasecmp($sp[1], "into"))
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: $query\n");
				unset($res);
				unset($sp);
				return FALSE;
			}
			$res[] = $sp[2];
			$iscol = $sp[3];

			if(!strcasecmp($iscol, "values"))
			{
				$res[] = 'values';
				foreach($sp as $l=>$v)
				{
					if($l < 5) continue;
					$p[] = $v;
				}
				$res[] = trim(join(" ", $p), " \t\n\r\0\x0B()");
				unset($p);
			}
			else
			{
				foreach($sp as $l=>$v)
				{
					if($l < 3) continue;
					$p[] = $v;
				}
				$tmp = trim(join(" ", $p), " \t\n\r\0\x0B()");
				$sp = preg_split("/\) values \(/i", $tmp);
				unset($tmp);

				if(count($sp) == 2)
				{
					$res[] = $sp[0];
					$res[] = $sp[1];
				}
				else
				{
					$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: $query\n");
					unset($res);
					unset($sp);
					return FALSE;
				}
			}
			unset($iscol);
		}
		else if(!strncasecmp($query, UPDATE_CMD,
				strlen(UPDATE_CMD)))
		{
			$sp = split(" ", $query);
			$res[] = $sp[0];
			$res[] = $sp[1];
			if(strcasecmp($sp[2], "set"))
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: $query\n");
				unset($res);
				unset($sp);
				return FALSE;
			}
			foreach($sp as $i=>$v)
			{
				if($i < 3) continue;
				$p[] = $v;
			}
			unset($sp);
			$tmp = trim(join(" ", $p));
			list($set, $where) = preg_split("/ where /i", $tmp);
			$res[] = $set;
			$res[] = $where;
			unset($set);
			unset($where);
			unset($tmp);
			unset($p);
		}
		else
		{
			$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax (unknown command): $query\n");
			unset($sp);
			unset($res);
			$res = FALSE;
		}

		return $res;
	}

	function &query($query)
	{
		$pquery = $this->split_query($query);
		$ret = FALSE;

		if($pquery)
		{
			$cmd = strtolower(current($pquery));
			next($pquery);
			$ret = $this->$cmd(&$pquery);
			unset($pquery);
		}

		return $ret;
	}

	function &create(&$queryarr)
	{
		$ret = FALSE;
		$cmd = strtolower(current($queryarr));
		next($queryarr);
		$cmd = "create_".$cmd;
		$ret = &$this->$cmd(&$queryarr);
		unset($queryarr);

		return $ret;
	}

	function &create_sequence(&$qarr)
	{
		$name = current($qarr);
		$start = next($qarr);
		$increment = next($qarr);
		$maxvalue = next($qarr);

		if(!isset($this->sequences[$name]))
		{
			$obj =& $this->dbh->sequence($name, "n");
			if($start) $obj->setval($start);
			else $obj->setval(1);

			$path = preg_replace("/\.seq$/i", "", $obj->table->path)."-seqdesc.xml";
			$tmp = &new SQLi_sequence(&$obj, $name, $path);
			$tmp->create_desc($start, $increment, $maxvalue);
			$this->sequences[$name] =& $tmp;
			unset($path);
		}
		else
		{
			$this->printerr(T_FUNCTION, __LINE__, "ERROR: Table '$table' already exists\n");
		}
	}

	function create_table(&$qarr)
	{
		$table = current($qarr);

		if(!isset($this->tables[$table]))
		{
			$desc = next($qarr);
			$t = &$this->dbh->table($table);
			unset($qarr);
			$path = preg_replace("/\.db$/i", "", $t->path)."-desc.xml";
			$obj = &new SQLi_table(&$t, $table, $path);
			$obj->create_desc($desc);
			unset($desc);
			$obj->load_desc();
			$this->tables[$table] = &$obj;
			unset($table);
			return TRUE;
		}
		else
		{
			$this->printerr(T_FUNCTION, __LINE__, "ERROR: Table '$table' already exists\n");
		}
	}

	function create_index(&$qarr)
	{
		$res = FALSE;
		$name = current($qarr);
		$table = next($qarr);
		$fields = next($qarr);

		if(isset($this->tables[$table]))
		{
			$missing = "";
			$dereg = "/( *)(\n*)(\r*)(\t*),( *)(\n*)(\r*)(\t*)/";
			$fa = preg_split($dereg, $fields);
			if(!$this->tables[$table]->fields_exists($fa, &$missing))
			{
				$this->tables[$table]->methods->buildindex($fa);
				$ix['INDEX']['attributes']['FIELDS'] =
								join(",", $fa);
				$ix['INDEX']['attributes']['NAME'] = $name;
				$idxs = $this->tables[$table]->xmldesc[0]['TABLE']['sub'][1]['INDICES']['sub'];
				for($i = 0 ; $i < count($idxs) ; $i++)
				{
					if($idxs[$i]['INDEX']['attributes']['NAME'] == $name)
					{
						$found = $i;
						break;
					}
				}

				if(isset($found))
					$this->tables[$table]->xmldesc[0]['TABLE']['sub'][1]['INDICES']['sub'][$i] = $ix;
				else
					$this->tables[$table]->xmldesc[0]['TABLE']['sub'][1]['INDICES']['sub'][] = $ix;
				unset($found);
				unset($i);
				unset($ix);
				unset($idxs);
			}
			else
			{
			$this->printerr(T_FUNCTION, __LINE__, "ERROR: Trying to create an index on the non existing fields '$missing' from the table '$table'\n");
			}
			unset($qarr);
			unset($fa);
			unset($missing);
			unset($dereg);
		}
		else
		{
			$this->printerr(T_FUNCTION, __LINE__, "ERROR: Trying to create an index on the non existing table '$table'\n");
		}

		return $res;
	}

	function create_database(&$querya)
	{
		$db = &new LDBI(current($querya), TRUE);
		$ret = !$db->error;
		$db->close();
		unset($db);
		unset($querya);
		return $ret;
	}

	function &select(&$queryarr)
	{
		$missing="";
		$fieldf = current($queryarr);
		$table = next($queryarr);
		$where = next($queryarr);
		$order = next($queryarr);

		if(!isset($this->tables[$table]))
		{
			$this->printerr(T_FUNCTION, __LINE__, "Table '$table' doesn't exist");
			return FALSE;
		}

		if($fieldf == "*") $fields = $this->tables[$table]->fields;
		else $fields = preg_split("/?( *),?( *)/", $fieldf,
					-1, PREG_SPLIT_NO_EMPTY);
		if(($fieldf != "*") &&
		$this->tables[$table]->fields_exists($fields, &$missing))
		{
			$this->printerr(T_FUNCTION, __LINE__, "$table as no fields named '$missing'");
			return FALSE;
		}

		if(strlen($where))
		{
			$whereclause = $this->parse_whereclause($where, $table);
			if(!is_array($whereclause))
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax in the WHERE clause");
				return FALSE;
			}
		}

		if(strlen($order))
		{
			$orderfields = preg_split("/?( *),?( *)/", $order,
					-1, PREG_SPLIT_NO_EMPTY);
			$missing = FALSE;
			if($this->tables[$table]->fields_exists($fields, &$missing))
			{
				$this->printerr(T_FUNCTION, __LINE__, "$table as no fields named '$missing'");
				return FALSE;
			}
		}

		$obj =& new SQLi_result(&$this->tables[$table]);
		if($obj && strlen($where))
		{
			foreach($whereclause as $wc)
			{
				$res[] = $this->tables[$table]->methods->search($wc);
			}

			foreach($res as $r)
			{
				foreach($r as $rec)
				{
					$tmp[] = $this->tables[$table]->methods->get($rec);
				}
			}
			$obj->add_result_set($tmp);

			return $obj;
		}
		else if($obj)
		{
			$res = $this->tables[$table]->methods->load();
			$obj->add_result_set($res);
			return $obj;
		}
		else
		{
			$this->printerr(T_FUNCTION, __LINE__, "FATAL ERROR: can't allocate the Result object");
		}
	}

	function update(&$qarr)
	{
		$wc = FALSE;
		$table = current($qarr);
		$set = next($qarr);
		$where = next($qarr);
		unset($qarr);
		$uq = $this->parse_updateset($set, $table);
		unset($set);
		for($i = 0 ; $uq && ($i < count($uq)) ; $i++)
		{
			list($key, $val) = each($uq[$i]);
			$u[$key] = $val;
		}
		unset($uq);

		if(strlen($where))
		{
			$wc = $this->parse_whereclause($where, $table);
			unset($where);
		}

		if(is_array($u) && !$wc)
		{
			$res = $this->tables[$table]->methods->update($u);
		}
		else if(is_array($u) && is_array($wc))
		{
			foreach($wc as $w)
			{
				$res += $this->tables[$table]->methods->update($u, $w);
			}
			unset($wc);
		}
		unset($table);

		return $res;
	}

	function insert(&$qarr)
	{
		$rc = FALSE;
		$cols = FALSE;

		$table = current($qarr);
		$values = next($qarr);
		if($values != 'values')
		{
			$dereg = "/( *)(\n*)(\r*)(\t*),( *)(\n*)(\r*)(\t*)/";
			$cols = preg_split($dereg, $values);
			unset($dereg);
		}

		$values = next($qarr);
		$f = $this->tables[$table]->fields_description;
		for($i = 0 ; ($i < count($f)) ; $i++)
		{
			if($cols)
			{
				if(preg_grep("/".preg_quote($f[$i]['NAME'])."/", $cols))
				{
					$rc[$f[$i]['NAME']] = $f[$i];
				}
				else $rc[$f[$i]['NAME']] = FALSE;
			}
			else
			{
				$rc[$f[$i]['NAME']] = $f[$i];
			}
		}

		if($this->parse_ins_values(&$rc, $values))
		{
			foreach($rc as $col=>$v)
			{
				$iv[$col] = $v['value'];
			}

			if(count($iv))
				return $this->tables[$table]->methods->insert($iv);
		}

		return FALSE;
	}

	function parse_ins_values(&$rc, $values)
	{
		$ri = 0;
		$inval = 1;
		$w = $inval;
		$start = FALSE;
		$vt=FALSE;

		$values = trim($values);
		$ptr = preg_split("//", $values);
		$res = array();

		for($i = 1 ; ($i < count($ptr)) && is_array($res); $i++)
		{
			switch($w)
			{
				case $inval:
					if(is_numeric($ptr[$i]) && !$vt)
					{
						$value .= $ptr[$i];
						$vt = "num";
					}
					else if($start && ($vt == 'str') &&
						($ptr[$i] == '\''))
					{
						$w = 0;
						$start = FALSE;
						$res[$ri]['v'] = $value;
						$res[$ri]['t'] = "s";
						$vt = FALSE;
						$value = FALSE;
						$start = FALSE;
					}
					else if(($ptr[$i] == '\'') &&
					!$start && !$vt)
					{
						$start = TRUE;
						$vt="str";
					}
					else if($start && ($vt=='str'))
					{
						$value .= $ptr[$i];
					}
					else if(is_string($ptr[$i]) && !$vt)
					{
						$fn = $ptr[$i];
						$vt="func";
					}
					else if($vt == 'func')
					{
						if(!$start)
						{
							if($ptr[$i] == '(')
							{
								$start = TRUE;
							}
							else $fn .= $ptr[$i];
						}
						else if($ptr[$i] == ')')
						{
							$start = FALSE;
							$vt = FALSE;
							$w = 0;
							$res[$ri]['f']= $fn;
							$res[$ri]['v']= $value;
							$res[$ri]['t']= "f";
							unset($value);
							unset($fn);
							$fn = FALSE;
							$value = FALSE;
						}
						else
						{
							$value .= $ptr[$i];
						}
					}
					else if((($ptr[$i] == ' ') ||
					($ptr[$i] == '\t') ||
					($ptr[$i] == ',') ||
					($ptr[$i] == '')) &&
					($vt == "num"))
					{
						if($ptr[$i] == ',')
						{
							$i--;
						}
						$w = 0;
						$res[$ri]['v'] = $value;
						$res[$ri]['t'] = "n";
						$vt=FALSE;
						$value=FALSE;
					}
					else if($start && ($vt == 'str') &&
						($ptr[$i] == '\\'))
					{
						$value .= $ptr[++$i];
					}
					else
					{
						$res = FALSE;
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax parsing value in: $where\n");
					}
					break;
				default:
					for(; ($i < count($ptr)) && preg_match("/[, ]/", $ptr[$i+1]) ; $i++);
					$ri++;
					$w = $inval;
			}
		}

		$i = 0;
		foreach($rc as $k=>$v)
		{
			if(!$v)
			{
				continue;
			}

			if(($v['TYPE'] == 'int') && ($res[$i]['t'] == 'n'))
			{
				$ret[$k] = $v;
				$ret[$k]['value'] = $res[$i]['v'];
			}
			else if(($v['TYPE'] == 'text') &&
				($res[$i]['t'] == 's'))
			{
				$ret[$k] = $v;
				$ret[$k]['value'] = $res[$i]['v'];
			}
			else if($res[$i]['t'] == 'f')
			{
				$ret[$k] = $v;
				$func = $res[$i]['f'];
				if(isset($this->sqlfunctions[$func]))
				{
					if($v['TYPE'] != $this->sqlfunctions[$func]['type'])
					{
						$this->printerr(T_FUNCTION, __LINE__,
						"'$k' requires ".$v['TYPE']." values");
						unset($rc);
						unset($ret);
						$ret = FALSE;
						$rc = FALSE;
						break;
					}

					$call = $this->sqlfunctions[$func]['call'];
					$tm = $this->$call(&$this, $res[$i]['v']);
					if(!isset($tm))
					{
						unset($ret);
						break;
					}
					$ret[$k]['value'] = $tm;
					unset($tm);
					unset($call);
				}
				else
				{
					$this->printerr(T_FUNCTION, __LINE__,
						"SQL Function '$func' does not exist");
					unset($ret);
					unset($rc);
					$rc = FALSE;
					$ret = FALSE;
					break;
				}
			}
			else
			{
				$this->printerr(T_FUNCTION, __LINE__,
					"'$k' requires ".$v['TYPE']." values");
				unset($ret);
				unset($rc);
				$rc = FALSE;
				$ret = FALSE;
				break;
			}
			$i++;
		}

		$rc = $ret;

		return $ret;
	}

	function sequence_next(&$obj, $seqname)
	{
		$ret = FALSE;
		if(isset($obj->sequences[$seqname]))
		{
			$s =& $obj->sequences[$seqname];
			$cur = $s->methods->current();

			if($s->maxvalue && (($cur->increment + $cur) > $s->maxvalue))
				$s->methods->setval($s->start);

			unset($cur);
			$ret = $s->methods->next($s->increment);
		}
		else
		{
			$this->printerr("nextval", __LINE__,
				"Sequence '$seqname' does not exist");
		}

		return $ret;
	}

	function sequence_curr(&$obj, $seqname)
	{
		$ret = FALSE;
		if(isset($obj->sequences[$seqname]))
		{
			$ret = $obj->sequences[$seqname]->methods->current();
		}
		else
		{
			$this->printerr("currval", __LINE__,
				"Sequence '$seqname' does not exist");
		}

		return $ret;
	}

	function parse_whereclause($where, $table)
	{
		$ri = 0;
		$inkey = 1;
		$inval = 2;
		$keych = "/[A-Z_]/i";
		$w = $inkey;
		$start = FALSE;
		$vt=FALSE;

		$where = trim($where);
		$ptr = preg_split("//", $where);
		$res = array();

		for($i = 1 ; ($i < count($ptr)) && is_array($res); $i++)
		{
			switch($w)
			{
				case $inkey:
					if(count($res) && (!$conj && !$key))
					{
						$res = FALSE;
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: $where");
						break;
					}
					else if(preg_match($keych, $ptr[$i]))
					{
						$conj=FALSE;
						$key .= $ptr[$i];
					}
					else if($key && ($ptr[$i] == '='))
					{
						$w=$inval;
						break;
					}
					else
					{
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax parsing key in: $where\n");
						$res = FALSE;
						break;
					}
					break;
				case $inval:
					if(is_numeric($ptr[$i]) && !$vt)
					{
						$value .= $ptr[$i];
						$vt = "num";
					}
					else if($start && ($vt == 'str') &&
						($ptr[$i] == '\''))
					{
						$w = 0;
						$start = FALSE;
						$res[$ri].=$key."=".$value." ";
						$check[][$key] = $value;
						$vt = FALSE;
						$key = FALSE;
						$value = FALSE;
						$start = FALSE;
					}
					else if(($ptr[$i] == '\'') &&
					!$start && !$vt)
					{
						$start = TRUE;
						$vt="str";
					}
					else if($start && ($vt=='str'))
					{
						$value .= $ptr[$i];
					}
					else if((($ptr[$i] == ' ') ||
					($ptr[$i] == '\t') ||
					($ptr[$i] == '')) &&
					($vt == "num"))
					{
						$w = 0;
						$res[$ri].=$key."=".$value." ";
						$check[][$key] = $value;
						$vt=FALSE;
						$key=FALSE;
						$value=FALSE;
					}
					else if($start && ($vt == 'str') &&
						($ptr[$i] == '\\'))
					{
						$value .= $ptr[++$i];
					}
					else
					{
						$res = FALSE;
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax parsing value in: $where\n");
					}
					break;
				default:
					if((($ptr[$i] == 'o') ||
						($ptr[$i] == 'O')) && !$start)
					{
						$start = 'O';
					}
					else if((($ptr[$i] == 'a') ||
						($ptr[$i] == 'A')) && !$start)
					{
						$start = 'A';
					}
					else if((($ptr[$i] == 'n') ||
						($ptr[$i] == 'N')) &&
						($start == 'A'))
					{
						$start = 'N';
					}
					else if((($ptr[$i] == 'd') ||
						($ptr[$i] == 'D')) &&
						($start == 'N'))
					{
						$res[$ri] .= "and ";
						$conj = TRUE;
						$start = FALSE;
						$stop = TRUE;
					}
					else if((($ptr[$i] == 'r') ||
						($ptr[$i] == 'R')) &&
						($start == 'O'))
					{
						$res[$ri] = substr($res[$ri], 0, -1);
						$ri++;
						$start = FALSE;
						$stop = TRUE;
					}
					else if($stop && (($ptr[$i] == ' ') ||
						($ptr[$i] == '\t')))
					{
						if(preg_match($keych,
							$ptr[$i+1]))
						{
							$w=$inkey;
							break;
						}
					}
					else if(!$stop && (($ptr[$i] == ' ') ||
						($ptr[$i] == '\t')))
					{
						break;
					}
					else if(!strlen($ptr[$i]) && !$start)
					{
						/* end of the string */
						break;
					}
					else
					{
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax parsing: $where\n");
						$res = FALSE;
						break;
					}
					break;
			}
		}

		$ex = array();
		for($i = 0 ; $check && ($i < count($check)) ; $i++)
		{
			$missing = FALSE;
			list($k, $val) = each($check[$i]);
			if(!array_key_exists($k, $ex))
			{
				$ex[$key] = TRUE;
				if($this->tables[$table]->fields_exists(array($key), &$missing))
				{
					$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: $table has no columns named '$key'\n");
					unset($res);
					unset($check);
					$res = FALSE;
					$check = FALSE;
				}
			}
			else
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: trying to set '$key' twice");
				unset($res);
				unset($check);
				$res = FALSE;
				$check = FALSE;
			}
			unset($k);
			unset($val);
		}

		unset($check);
		unset($ex);
		unset($where);
		unset($table);

		return $res;
	}

	function parse_updateset($set, $table)
	{
		$ri = 0;
		$inkey = 1;
		$inval = 2;
		$keych = "/[A-Z_]/i";
		$w = $inkey;
		$start = FALSE;
		$vt=FALSE;

		$set = trim($set);
		$ptr = preg_split("//", $set);
		$res = array();

		for($i = 1 ; ($i < count($ptr)) && is_array($res); $i++)
		{
			switch($w)
			{
				case $inkey:
					if(count($res) && (!$conj && !$key))
					{
						$res = FALSE;
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: $set");
						break;
					}
					else if(preg_match($keych, $ptr[$i]))
					{
						$conj=FALSE;
						$key .= $ptr[$i];
					}
					else if($key && ($ptr[$i] == '='))
					{
						$w=$inval;
						break;
					}
					else
					{
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax parsing key in: $set\n");
						$res = FALSE;
						break;
					}
					break;
				case $inval:
					if(is_numeric($ptr[$i]) && !$vt)
					{
						$value .= $ptr[$i];
						$vt = "num";
					}
					else if($start && ($vt == 'str') &&
						($ptr[$i] == '\''))
					{
						$w = 0;
						$start = FALSE;
						$res[$ri][$key] = $value;
						$check[][$key] = $value;
						$vt = FALSE;
						$key = FALSE;
						$value = FALSE;
						$start = FALSE;
					}
					else if(($ptr[$i] == '\'') &&
					!$start && !$vt)
					{
						$start = TRUE;
						$vt="str";
					}
					else if($start && ($vt=='str'))
					{
						$value .= $ptr[$i];
					}
					else if((($ptr[$i] == ' ') ||
					($ptr[$i] == '') ||
					($ptr[$i] == '\t')) &&
					($vt == "num"))
					{
						$w = 0;
						$res[$ri][$key] = $value;
						$check[][$key] = $value;
						$vt=FALSE;
						$key=FALSE;
						$value=FALSE;
					}
					else if($start && ($vt == 'str') &&
						($ptr[$i] == '\\'))
					{
						$value .= $ptr[++$i];
					}
					else if(is_string($ptr[$i]) && !$vt)
					{
						$fn = $ptr[$i];
						$vt="func";
					}
					else if($vt == 'func')
					{
						if(!$start)
						{
							if($ptr[$i] == '(')
							{
								$start = TRUE;
							}
							else $fn .= $ptr[$i];
						}
						else if($ptr[$i] == ')')
						{
							$start = FALSE;
							$vt = FALSE;
							$w = 0;
							$f['f']= $fn;
							$f['v']= $value;
							$res[$ri][$key] = $f;
							unset($key);
							unset($value);
							unset($fn);
							$fn = FALSE;
							$value = FALSE;
							$key = FALSE;
						}
						else
						{
							$value .= $ptr[$i];
						}
					}
					else
					{
						$res = FALSE;
						$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax parsing value in: '$set' \n");
					}
					break;
				default:
					for(; ($i < count($ptr)) && !preg_match("/[^, ]/", $ptr[$i+1]) ; $i++);
					$ri++;
			}
		}

		$ex = array();
		for($i = 0 ; $check && ($i < count($check)) ; $i++)
		{
			$missing = FALSE;
			list($k, $val) = each($check[$i]);
			if(!array_key_exists($k, $ex))
			{
				$ex[$key] = TRUE;
				if($this->tables[$table]->fields_exists(array($key), &$missing))
				{
					$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: $table has no columns named '$key'\n");
					$res = FALSE;
					$check = FALSE;
				}
			}
			else
			{
				$this->printerr(T_FUNCTION, __LINE__, "Bad Syntax: trying to set '$key' twice");
				$res = FALSE;
				$check = FALSE;
			}
		}

		for($i = 0 ; $i < count($res) ; $i++)
		{
			list($k, $v) = each($res[$i]);

			if(is_array($v))
			{
				$call = $this->sqlfunctions[$v['f']]['call'];
				$tm = $this->$call(&$this, $v['v']);
				if(!isset($tmp))
				{
					unset($ret);
					break;
				}
				$v = $tm;
			}
			$ret[$i][$k] = $v;
		}
		unset($res);

		return $ret;
	}

	function close()
	{
		$this->dbh->close();
		$this->save_table_config();
		$this->save_sequence_config();
		$this->rebuild_indices();
	}

	function flush()
	{
		$this->save_table_config();
		$this->save_sequence_config();
		$this->rebuild_indices();
	}

	function rebuild_indices()
	{
		foreach($this->tables as $obj)
		{
			$obj->rebuild_indices();
		}
		reset($this->tables);
	}

	function load_table_config($tobj)
	{
		$name = $tobj->name;
		$path = preg_replace("/\.db$/i", "", $tobj->path)."-desc.xml";
		$this->tables[$name] = &new SQLi_table(&$tobj, $name, $path);
		$this->tables[$name]->load_desc();
	}

	function load_sequence_config($sobj)
	{
		$name = $sobj->name;
		$path = preg_replace("/\.seq$/i", "", $sobj->path)."-seqdesc.xml";
		$this->sequences[$name] = &new SQLi_sequence(&$sobj, $name, $path);
		$this->sequences[$name]->load_desc();
	}

	function save_table_config()
	{
		if(is_array($this->tables))
		{
			foreach($this->tables as $table=>$obj)
			{
				$obj->save_config();
			}
		}
	}

	function save_sequence_config()
	{
		if(is_array($this->sequences))
		{
			foreach($this->sequences as $seq=>$obj)
			{
				$obj->save_config();
			}
		}
	}

	function printerr($f, $l, $str)
	{
		echo "$f:$l - $str\n";
	}
}

class SQLi_table
{
	function SQLi_table(&$table, $name, $path)
	{
		$this->name = $name;
		$this->path = $path;
		$this->methods =& $table;
	}

	function create_desc($desc)
	{
		$dereg = "/( *)(\n*)(\r*)(\t*),( *)(\n*)(\r*)(\t*)/";
		$fields = preg_split($dereg, $desc);
		unset($dereg);
		foreach($fields as $fieldd)
		{
			list($field, $type) = split(" ", $fieldd);
			$tmp['FIELD']['attributes']['NAME'] = $field;
			$tmp['FIELD']['attributes']['TYPE'] = $type;
			$xfdesc[] = $tmp;
			unset($tmp);
			unset($fieldd);
		}
		unset($fields);
		$xdesc[0]['TABLE']['attributes']['NAME'] = $this->name;
		$xdesc['NAME'] = 'DESCRIPTION';

		$xdesc[0]['TABLE']['sub'][0]['FIELDS']['sub'] = $xfdesc;
		unset($xfdesc);
		$xml =& new apk_xml(APK_XML_OUTTYPE_FILE_TRUNCATE, $this->path);
		$xml->db = $xdesc;
		unset($xdesc);
		$xml->dump();
		unset($xml);
	}

	function load_desc()
	{
		$xml =& new apk_xml(APK_XML_OUTTYPE_BUFFER);
		$xml->parse_file($this->path);
		$desc = $xml->build_struct();
		$this->parse_fields($desc);
		$this->parse_indices($desc);
		$this->xmldesc = $desc;
		unset($xml);
		unset($desc);
	}

	function parse_fields($desc)
	{
		$flist = $desc[0]['TABLE']['sub'][0]['FIELDS']['sub'];
		for($i = 0 ; $i < count($flist) ; $i++)
		{
			$this->fields[] = $flist[$i]['FIELD']['attributes']['NAME'];
			$this->fields_description[] = $flist[$i]['FIELD']['attributes'];
		}
		unset($i);
		unset($flist);
	}

	function parse_indices($desc)
	{
		$flist = $desc[0]['TABLE']['sub'][1]['INDICES']['sub'];
		for($i = 0 ; $i < count($flist) ; $i++)
		{
			$ifields = $flist[$i]['INDEX']['attributes']['FIELDS'];
			$tmp['name'] = $flist[$i]['INDEX']['attributes']['NAME'];
			$tmp['fields'] = split(",", $ifields);
			$this->indices[] = $tmp;
			unset($tmp);
			unset($ifields);
		}
		unset($i);
	}

	function fields_exists($check, &$missing)
	{
		$ret = TRUE;
		for($i = 0 ; $i < count($check) ; $i++)
		{
			$in = TRUE;
			list($f) = preg_grep("/^".preg_quote($check[$i])."$/", 
						$this->fields);
			if(!$f)
			{
				$ret = FALSE;
				$missing .= $check[$i]." ";
			}
		}

		if(!$in) $ret = FALSE;
		unset($in);
		unset($i);

		return $ret;
	}

	function save_config()
	{
		list($usec, $sec) = explode(' ', microtime());
		mt_srand((float) $sec + ((float) $usec * 100000));
		$r = mt_rand();
		$xml = &new apk_xml(APK_XML_OUTTYPE_FILE_TRUNCATE,
				$this->path.".$r.tmp");
		$xml->db = $this->xmldesc;
		$xml->dump();
		unset($xml);
		rename($this->path.".$r.tmp", $this->path);
		unset($r);
		$cpath = dirname($this->path)."/.".basename($this->path).".cache";
		unlink($cpath);
		unset($cpath);
		$this->load_desc();
	}

	function rebuild_indices()
	{
		for($i = 0 ; $i < count($this->indices) ; $i++)
		{
			$this->methods->buildindex($this->indices[$i]['fields']);
		}
	}
}

class SQLi_sequence
{
	function SQLi_sequence(&$seq, $name, $path)
	{
		$this->name = $name;
		$this->path = $path;
		$this->methods =& $seq;
	}

	function create_desc($start, $increment, $maxval)
	{
		$this->start = $start;
		$this->increment = $increment;
		$this->maxvalue = $maxval;
		$tmp['attributes']['START'] = $start;
		$tmp['attributes']['INCREMENT'] = $increment;
		$tmp['attributes']['MAXVALUE'] = $maxval;
		$tmp['attributes']['NAME'] = &$this->name;
		$xdesc[0]['SEQUENCE'] = &$tmp;
		$xdesc['NAME'] = 'DESCRIPTION';

		$xml =& new apk_xml(APK_XML_OUTTYPE_FILE_TRUNCATE, $this->path);
		$this->xmldesc = $xml->db = $xdesc;
		$xml->dump();
		unset($xdesc);
		unset($xml);
	}

	function load_desc()
	{
		$xml =& new apk_xml(APK_XML_OUTTYPE_BUFFER);
		$xml->parse_file($this->path);
		$xml->build_struct();
		$this->xmldesc = $xml->db;
		$this->start = $this->xmldesc[0]['SEQUENCE']['attributes']['START'];
		$this->increment = $this->xmldesc[0]['SEQUENCE']['attributes']['INCREMENT'];
		$this->maxvalue = $this->xmldesc[0]['SEQUENCE']['attributes']['MAXVALUE'];
		unset($xml);
	}

	function save_config()
	{
		list($usec, $sec) = explode(' ', microtime());
		mt_srand((float) $sec + ((float) $usec * 100000));
		$r = mt_rand();
		$xml = &new apk_xml(APK_XML_OUTTYPE_FILE_TRUNCATE,
				$this->path.".$r.tmp");
		$xml->db = $this->xmldesc;
		$xml->dump();
		unset($xml);
		rename($this->path.".$r.tmp", $this->path);
		unset($r);
		$cpath = dirname($this->path)."/.".basename($this->path).".cache";
		unlink($cpath);
		unset($cpath);
		$this->load_desc();
	}
}

class SQLi_result
{
	function SQLi_result(&$parent_table_ptr)
	{
		$this->parent =& $parent_table_ptr;
	}

	function add_result_set($res)
	{
		if(is_array($res))
		foreach($res as $rec)
		{
			$this->result[] = $rec['d'];
		}
	}
}

?>
